from __future__ import annotations

from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional

from mayim.base.interface import BaseInterface
from mayim.exception import MayimError

try:
    from asyncmy import Connection, create_pool

    MYSQL_ENABLED = True
except ModuleNotFoundError:
    MYSQL_ENABLED = False
    Connection = type("Connection", (), {})  # type: ignore


class MysqlPool(BaseInterface):
    """Interface for connecting to a MySQL database"""

    scheme = "mysql"

    def _setup_pool(self):
        if not MYSQL_ENABLED:
            raise MayimError(
                "MySQL driver not found. Try reinstalling Mayim: "
                "pip install mayim[mysql]"
            )
        self._pool = create_pool(
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            db=self.db,
        )

    async def open(self):
        """Open connections to the pool"""
        self._pool = await self._pool

    async def close(self):
        """Close connections to the pool"""
        self._pool.close()
        await self._pool.wait_closed()

    @asynccontextmanager
    async def connection(
        self, timeout: Optional[float] = None
    ) -> AsyncIterator[Connection]:
        """Obtain a connection to the database

        Args:
            timeout (float, optional): _Not implemented_. Defaults to `None`.

        Returns:
            AsyncIterator[Connection]: Iterator that will yield a connection

        Yields:
            Iterator[AsyncIterator[Connection]]: A database connection
        """
        existing = self.existing_connection()
        if existing:
            yield existing
        else:
            transaction = self.in_transaction()
            async with self._pool.acquire() as conn:
                if transaction:
                    await conn.begin()
                yield conn
                if transaction:
                    if self.do_commit():
                        await conn.commit()



from typing import Optional, Type

from mayim.base.interface import BaseInterface
from mayim.exception import MayimError


class LazyPool(BaseInterface):
    _singleton = None
    _derivative: Optional[Type[BaseInterface]]

    def __new__(cls, *args, **kwargs):
        if cls._singleton is None:
            cls._singleton = super().__new__(cls)
            cls._singleton._derivative = None
            cls._singleton._derivative_dsn = ""
        return cls._singleton

    def _setup_pool(self):
        ...

    def _populate_dsn(self):
        ...

    def _populate_connection_args(self):
        ...

    async def open(self):
        ...

    async def close(self):
        ...

    def connection(self, timeout: Optional[float] = None):
        ...

    def set_derivative(self, interface_class: Type[BaseInterface]) -> None:
        self._derivative = interface_class

    def set_dsn(self, dsn: str) -> None:
        self._derivative_dsn = dsn

    def derive(self) -> BaseInterface:
        if not self._derivative:
            raise MayimError("No interface available to derive")
        return self._derivative(dsn=self._derivative_dsn)



from __future__ import annotations

from sqlite3 import Cursor
from typing import Any, Dict, Optional, Sequence, Tuple

from mayim.sql.sqlite.query import SQLiteQuery

from ..executor import SQLExecutor

try:
    import aiosqlite  # noqa

    AIOSQLITE_ENABLED = True
except ModuleNotFoundError:
    AIOSQLITE_ENABLED = False


class SQLiteExecutor(SQLExecutor):
    """Executor for interfacing with a SQLite database"""

    ENABLED = AIOSQLITE_ENABLED
    QUERY_CLASS = SQLiteQuery
    POSITIONAL_SUB = r"?"
    KEYWORD_SUB = r":\2"

    async def _run_sql(
        self,
        query: str,
        name: str = "",
        as_list: bool = False,
        no_result: bool = False,
        posargs: Optional[Sequence[Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        method_name = self._get_method(as_list=as_list)
        async with self.pool.connection() as conn:
            exec_values = list(posargs) if posargs else params
            conn.row_factory = self._dict_factory
            cursor = await conn.execute(query, exec_values)
            if no_result:
                return None
            raw = await getattr(cursor, method_name)()
            return raw

    @staticmethod
    def _dict_factory(cursor: Cursor, row: Tuple[Any, ...]) -> Dict[str, Any]:
        return {val[0]: row[idx] for idx, val in enumerate(cursor.description)}



import re

from mayim.exception import MayimError
from mayim.sql.query import ParamType, SQLQuery


class SQLiteQuery(SQLQuery):
    __slots__ = ("name", "text", "param_type")
    PATTERN_POSITIONAL_PARAMETER = re.compile(r"\?")
    PATTERN_KEYWORD_PARAMETER = re.compile(r"\:[a-z_][a-z0-9_]")

    def __init__(self, name: str, text: str) -> None:
        super().__init__(name, text)
        positional_argument_exists = bool(
            self.PATTERN_POSITIONAL_PARAMETER.search(self.text)
        )
        keyword_argument_exists = bool(
            self.PATTERN_KEYWORD_PARAMETER.search(self.text)
        )

        if positional_argument_exists and keyword_argument_exists:
            raise MayimError("Positional and keyword arguments provided")
        if positional_argument_exists:
            self.param_type = ParamType.POSITIONAL
        elif keyword_argument_exists:
            self.param_type = ParamType.KEYWORD
        else:
            self.param_type = ParamType.NONE



from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional

from mayim.base.interface import BaseInterface
from mayim.exception import MayimError

try:
    from psycopg import AsyncConnection
    from psycopg_pool import AsyncConnectionPool

    POSTGRES_ENABLED = True
except ModuleNotFoundError:
    POSTGRES_ENABLED = False
    AsyncConnection = type("Connection", (), {})  # type: ignore
    AsyncConnectionPool = type("Connection", (), {})  # type: ignore


class PostgresPool(BaseInterface):
    """Interface for connecting to a Postgres database"""

    scheme = "postgres"

    def _setup_pool(self):
        if not POSTGRES_ENABLED:
            raise MayimError(
                "Postgres driver not found. Try reinstalling Mayim: "
                "pip install mayim[postgres]"
            )
        self._pool = AsyncConnectionPool(self.full_dsn)

    async def open(self):
        """Open connections to the pool"""
        await self._pool.open()

    async def close(self):
        """Close connections to the pool"""
        await self._pool.close()

    @asynccontextmanager
    async def connection(
        self, timeout: Optional[float] = None
    ) -> AsyncIterator[AsyncConnection]:
        """Obtain a connection to the database

        Args:
            timeout (float, optional): Time before an error is raised on
                failure to connect. Defaults to `None`.

        Returns:
            AsyncIterator[Connection]: Iterator that will yield a connection

        Yields:
            Iterator[AsyncIterator[Connection]]: A database connection
        """
        existing = self._connection.get(None)
        if existing:
            yield existing
        else:
            async with self._pool.connection(timeout=timeout) as conn:
                yield conn



from asyncio import get_running_loop
from inspect import isclass
from typing import Optional, Sequence, Type, TypeVar, Union
from urllib.parse import urlparse

from mayim.base import Executor, Hydrator
from mayim.base.interface import BaseInterface
from mayim.exception import MayimError
from mayim.lazy.interface import LazyPool
from mayim.registry import InterfaceRegistry, Registry
from mayim.sql.postgres.interface import PostgresPool

T = TypeVar("T", bound=Executor)
DEFAULT_INTERFACE = PostgresPool


class Mayim:
    """Main entryway for initializing access to the data layer.

    Where possible, it is advised (although not necessary) to initialize the
    Mayim object once the async loop is running.

    Example:

    ```python
    async def run():
        Mayim(
            executors=[MyExecutor]
            dsn="postgres://user:pass@localhost:5432/mydb"
        )
    ```
    """

    def __init__(
        self,
        *,
        executors: Optional[Sequence[Union[Type[Executor], Executor]]] = None,
        dsn: str = "",
        db_path: str = "",
        hydrator: Optional[Hydrator] = None,
        pool: Optional[BaseInterface] = None,
        strict: bool = True,
    ):
        """Initializer for Mayim instance

        The `dsn`, `db_path`, and the `pool` are mutually exclusive. Either
        one or none of them should be passed. If one exists, it will be used
        to create the fallback interface pool for the datasource. Therefore,
        if an executor does not define its own, then it will use the fallback
        pool as established here.

        Similarly, the `hydrator` is used in the case that an executor does
        not define a hydrator. If one is not provided, then a generic hydrator
        instance will be created.

        Executors need not be registered when the `Mayim` instance is
        instantiated. See
        [load](./mayim.mayim.html#load) and [loading and instantiating](/guide/executors.html#loading-and-instantiating)
        for more details.

        Args:
            executors (Sequence[Union[Type[Executor], Executor]], optional): Executors
                that are being loaded at initialization time.
                Defaults to `None`.
            dsn (str, optional): DSN to the data source. Defaults to `""`.
            db_path (str, optional): Path to a SQLite database.
                Defaults to `""`.
            hydrator (Hydrator, optional): Fallback hydrator to use if not
                specified. Defaults to `None`.
            pool (BaseInterface, optional): Fallback database interface.
                Defaults to `None`.
            strict (bool, optional): Whether to raise an error if there is
                an empty method in a decorator but no loaded query.
                Defaults to `True`.

        Raises:
            MayimError: If there is conflicing data access source
        """  # noqa
        if pool and dsn:
            raise MayimError("Conflict with pool and DSN")

        if dsn and db_path:
            raise MayimError("Conflict with DSN and DB path")

        if db_path and not dsn:
            dsn = db_path

        if not pool and dsn:
            pool_type = self._get_pool_type(dsn)
            try:
                get_running_loop()
            except RuntimeError:
                pool = LazyPool()
                pool.set_derivative(pool_type)
                pool.set_dsn(dsn)
            else:
                pool = pool_type(dsn)

        if not executors:
            executors = []
        self.strict = strict
        self.load(executors=executors, hydrator=hydrator, pool=pool)

        if hydrator is None:
            hydrator = Hydrator()

        Executor._fallback_hydrator = hydrator
        Executor._fallback_pool = pool

    @staticmethod
    def get(executor: Type[T]) -> T:
        """Fetch an instance of an executor

        This is useful when there you wish to retrieve the registered instance
        of an executor. Can be used in various parts of
        an application without having to pass variables.

        Example:

        ```python
        from mayim import Mayim
        from my.package.executors import AwesomeExecutor

        async def some_func():
            executor = Mayim.get(AwesomeExecutor)
            awesomeness = await executor.select_something_awesome()
            ...
        ```

        Args:
            executor (Type[T]): The class of the registered executor instance

        Raises:
            MayimError: If the passed executor has not been registered

        Returns:
            Executor: The executor instance
        """
        registry = Registry()
        try:
            instance = registry[executor.__name__]
        except KeyError as e:
            raise MayimError(f"{executor} has not been registered") from e
        return instance

    def register(cls, executor: Executor) -> None:
        """Register an Executor instance

        Args:
            executor (Executor): The instance to be registered
        """
        Registry().register(executor)

    def reset_registry(self):
        """Empty the executor registry"""
        Registry.reset()

    def load(
        self,
        *,
        executors: Sequence[Union[Type[Executor], Executor]],
        hydrator: Optional[Hydrator] = None,
        pool: Optional[BaseInterface] = None,
        strict: Optional[bool] = None,
    ) -> None:
        """Look through the Executor definition for methods that should
        execute a query, then load the corresponding SQL from a .sql file
        and hold in memory.

        Args:
            executors (Sequence[Union[Type[Executor], Executor]]): The
                executor classes or instances to be registered
            hydrator (Hydrator, optional): Fallback hydrator to use if not
                specified. Defaults to `None`.
            pool (BaseInterface, optional): Fallback database interface.
                Defaults to `None`.
            strict (bool, optional): Whether to raise an error if there is
                an empty method in a decorator but no loaded query.
                Defaults to `True`.

        Raises:
            MayimError: If there is no fallback database interface, Mayim will
                raise an exception if there is a registered executor without
                an instance-specific database interface
        """
        to_load = set(executors)
        to_load.update(Registry().values())
        strict = strict if strict is not None else self.strict
        for executor in to_load:
            if isclass(executor):
                try:
                    executor = self.get(executor)
                except MayimError:
                    ...
            if isinstance(executor, Executor):
                if executor.pool is LazyPool():
                    if not pool:
                        raise MayimError(
                            f"Cannot load {executor} without a pool"
                        )
                    executor._pool = pool
                executor = executor.__class__
            else:
                executor(pool=pool, hydrator=hydrator)

            if executor._loaded:
                continue

            executor._load(strict)

    def _get_pool_type(self, dsn: str) -> Type[BaseInterface]:
        parts = urlparse(dsn)

        for interface_type in BaseInterface.registered_interfaces:
            if parts.scheme == interface_type.scheme:
                return interface_type
        return DEFAULT_INTERFACE

    async def connect(self) -> None:
        """Connect to all database interfaces"""
        registry = Registry()
        to_derive = {
            executor
            for executor in registry.values()
            if isinstance(executor.pool, LazyPool)
        }
        for executor in to_derive:
            derived = executor.pool.derive()
            executor._pool = derived
            if isinstance(executor.__class__._fallback_pool, LazyPool):
                executor.__class__._fallback_pool = derived

        for interface in InterfaceRegistry():
            await interface.open()

    async def disconnect(self) -> None:
        """Disconnect from all database interfaces"""
        for interface in InterfaceRegistry():
            await interface.close()



from inspect import Parameter
from typing import Any, Dict, List, Type, Union


class Hydrator:
    """Object responsible for casting from the data layer to a model"""

    fallback: Type[object] = dict
    """The model type that will be used if there is none passed in the
    hydrate method"""

    def _make(self, model: Type[object]):
        def factory(data: Union[Dict[str, Any], List[Dict[str, Any]]]):
            if model is None:
                return None
            if isinstance(data, list):
                return [self.hydrate(row, model=model) for row in data]
            return self.hydrate(data, model=model)

        return factory

    def hydrate(
        self, data: Dict[str, Any], model: Type[object] = Parameter.empty
    ):
        """Perform casting operation

        Args:
            data (Dict[str, Any]): Raw data from the source
            model (Type[object], optional): The model that will do the
                casting. If no value is passed, it will use whatever the
                Hydrator's fallback value is set to. Defaults to
                `Parameter.empty`.

        Returns:
            _type_: The data cast into the model
        """
        if model is Parameter.empty:
            model = self.fallback
        elif model in (str, int, float, bool):
            return model(*data.values())
        elif model.__name__ == "Dict":
            return data
        return model(**data)



from typing import Optional, Sequence, Type, Union

from mayim import Executor, Hydrator, Mayim
from mayim.base.interface import BaseInterface
from mayim.exception import MayimError
from mayim.extension.statistics import (
    display_statistics,
    log_statistics_report,
    setup_query_counter,
)
from mayim.registry import InterfaceRegistry, Registry

try:
    from sanic.helpers import Default, _default
    from sanic.log import logger
    from sanic.signals import Event
    from sanic_ext import Extend
    from sanic_ext.extensions.base import Extension

    SANIC_INSTALLED = True
except ModuleNotFoundError:
    SANIC_INSTALLED = False
    Extension = type("Extension", (), {})  # type: ignore
    Extend = type("Extend", (), {})  # type: ignore


class SanicMayimExtension(Extension):
    name = "mayim"

    def __init__(
        self,
        *,
        executors: Optional[Sequence[Union[Type[Executor], Executor]]] = None,
        dsn: str = "",
        hydrator: Optional[Hydrator] = None,
        pool: Optional[BaseInterface] = None,
        counters: Union[Default, bool] = _default,
    ):
        if not SANIC_INSTALLED:
            raise MayimError(
                "Could not locate either Sanic or Sanic Extensions. "
                "Both libraries must be installed to use SanicMayimExtension. "
                "Try: pip install sanic[ext]"
            )
        self.executors = executors or []
        for executor in self.executors:
            Registry().register(executor)
        self.mayim_kwargs = {
            "dsn": dsn,
            "hydrator": hydrator,
            "pool": pool,
        }
        self.counters = counters

    def startup(self, bootstrap: Extend) -> None:
        for executor in Registry().values():
            if isinstance(executor, Executor):
                bootstrap.dependency(executor)
            else:
                bootstrap.add_dependency(
                    executor, lambda *_: Mayim.get(executor)
                )

        @self.app.before_server_start
        async def setup(_):
            Mayim(executors=self.executors, **self.mayim_kwargs)
            for interface in InterfaceRegistry():
                logger.info(f"Opening {interface}")
                await interface.open()

        @self.app.after_server_stop
        async def shutdown(_):
            for interface in InterfaceRegistry():
                logger.info(f"Closing {interface}")
                await interface.close()

        for executor in Registry().values():
            if isinstance(executor, Executor):
                bootstrap.dependency(executor)
            else:
                bootstrap.add_dependency(
                    executor, lambda *_: Mayim.get(executor)
                )
        if display_statistics(self.counters, self.executors):
            self.app.signal(Event.HTTP_LIFECYCLE_REQUEST)(setup_query_counter)

        if display_statistics(self.counters, self.executors):
            self.app.on_request(setup_query_counter)

            @self.app.on_response
            async def display(*_):
                log_statistics_report(logger)

    def render_label(self):
        length = len(Registry())
        s = "" if length == 1 else "s"
        return f"[{length} executor{s}]"



from __future__ import annotations

from ast import AsyncFunctionDef, Constant, Expr, FunctionDef, Pass, parse
from contextvars import ContextVar
from inspect import cleandoc, getdoc, getmodule, getsource, stack
from pathlib import Path
from textwrap import dedent
from typing import (
    Any,
    Dict,
    Generic,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from mayim.base.hydrator import Hydrator
from mayim.base.interface import BaseInterface
from mayim.base.query import Query
from mayim.exception import MayimError
from mayim.lazy.interface import LazyPool
from mayim.registry import Registry

T = TypeVar("T", bound=Query)


class Executor(Generic[T]):
    """
    Base class for creating executors, which serve as the main interface for
    interacting with a data source. Likely you will want to create a subclass
    from one of its subclasses and not directly from this base class.
    """

    _queries: Dict[str, T]
    _fallback_hydrator: Hydrator
    _fallback_pool: Optional[BaseInterface]
    _loaded: bool = False
    _hydrators: Dict[str, Hydrator]
    path: Optional[Union[str, Path]] = None
    """`Optional[Union[str, Path]]`: Class property that is a custom  path "
        location of queries to be loaded. Default to `None`"""
    ENABLED: bool = True
    QUERY_CLASS: Type[T]

    def __init__(
        self,
        pool: Optional[BaseInterface] = None,
        hydrator: Optional[Hydrator] = None,
    ) -> None:
        """Base class for creating executors

        Args:
            pool (BaseInterface, optional): An interface used
                for a specific executor to override a global pool.
                Defaults to `None`.
            hydrator (Hydrator, optional): A hydrator used
                for a specific executor to override a global hydrator.
                Defaults to `None`.

        Raises:
            MayimError: If a dependency is missing
        """
        if not self.ENABLED:
            raise MayimError(
                f"Cannot instantiate {self.__class__.__name__}. "
                "Perhaps you have a missing dependency?"
            )
        pool = pool or getattr(self.__class__, "_fallback_pool", None)
        if not pool:
            pool = LazyPool()
        self._pool = pool
        self._hydrator = hydrator
        self._context: ContextVar[Tuple[Type[object], str]] = ContextVar(
            "_context"
        )
        Registry().register(self)

    @property
    def hydrator(self) -> Hydrator:
        """The assigned hydrator. Will return an instance specific hydrator
        if one was assigned.

        Returns:
            Hydrator: The hydrator
        """
        if self._hydrator:
            return self._hydrator
        return self._fallback_hydrator

    @property
    def pool(self) -> BaseInterface:
        """The assigned pool. Will return an instance specific pool
        if one was assigned.

        Returns:
            BaseInterface: The pool interface
        """
        return self._pool

    def execute(
        self,
        query: Union[str, Query],
        name: str = "",
        model: Optional[Type[object]] = None,
        as_list: bool = False,
        allow_none: bool = False,
        posargs: Optional[Sequence[Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Low-level API to execute a query and hydrate the results

        Args:
            query (Union[str, Query]): The query to be executed
            name (str, optional): The name of the query. Defaults to `""`.
            model (Type[object], optional): The model to be used
                for hydration. Defaults to `None`.
            as_list (bool, optional): Whether to return the results as a
                list of hydrated objects. Defaults to `False`.
            allow_none (bool, optional): Whether `None` is an acceptable return
                value. Defaults to `False`.
            posargs (Sequence[Any], optional): Positional arguments.
                Defaults to `None`.
            params (Dict[str, Any], optional): Keyword arguments.
                Defaults to `None`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not define execute"
        )

    def run_sql(
        self,
        query: str = "",
        name: str = "",
        as_list: bool = False,
        no_result: bool = False,
        posargs: Optional[Sequence[Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Low-level API to execute a query and return the results without
        hydration.

        Args:
            query (Union[str, Query]): The query to be executed
            name (str, optional): The name of the query. Defaults to `""`.
            as_list (bool, optional): Whether to return the results as a
                list of hydrated objects. Defaults to `False`.
            allow_none (bool, optional): Whether `None` is an acceptable return
                value. Defaults to `False`.
            posargs (Sequence[Any], optional): Positional arguments.
                Defaults to `None`.
            params (Dict[str, Any], optional): Keyword arguments.
                Defaults to `None`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not define run_sql"
        )

    def get_query(self, name: Optional[str] = None) -> T:
        """Return a query

        Args:
            name (str, optional): The name of the query to be
                retrieved. If no name is supplied, it will look for the query
                by the name of the calling method. Defaults to `None`.

        Raises:
            MayimError: When the query coul not be found

        Returns:
            Query: A query object
        """
        if not name:
            for frame in stack():
                if self.is_query_name(frame.function):
                    name = frame.function
                    break
            if not name:
                raise MayimError(
                    "Could not find query. Please specify a name."
                )
        return self._queries[name]

    def get_hydrator(self, name: Optional[str] = None) -> Hydrator:
        """Return a hydrator

        Args:
            name (str, optional): The name of the hydrator to be
                retrieved. If no name is supplied, it will look for the
                hydrator by the name of the calling method. Defaults to `None`.

        Raises:
            MayimError: When the hydrator coul not be found

        Returns:
            Hydrator: A hydrator object
        """
        if not name:
            for frame in stack():
                if self.is_query_name(frame.function):
                    name = frame.function
                    break
            if not name:
                raise MayimError(
                    "Could not find hydrator. Please specify a name."
                )
        return self._hydrators.get(name, self.hydrator)

    @classmethod
    def _load(cls, strict: bool) -> None:
        ...

    @staticmethod
    def is_query_name(obj) -> bool:
        """Whether an object is a valid query name

        Args:
            obj (Any): Depends upon which subclass is being implemented

        Returns:
            bool: is it valid
        """
        return False

    @staticmethod
    def _setup(func):
        return func

    @classmethod
    def get_base_path(cls, directory_name: Optional[str]) -> Path:
        """Get the base path for where queries will be located

        Args:
            directory_name (str, optional): A starting directory

        Raises:
            MayimError: When a module does not exist

        Returns:
            Path: The base path
        """
        module = getmodule(cls)
        if not module or not module.__file__:
            raise MayimError(f"Could not locate module for {cls}")

        base = Path(module.__file__).parent
        if cls.path is not None:
            if isinstance(cls.path, Path):
                return cls.path
            else:
                # TODO:
                # - support absolute path strings
                directory_name = cls.path
        if directory_name:
            base = base / directory_name
        return base


def is_auto_exec(func) -> bool:
    """Check if a method should be auto-executed.

    Example:

    Any of the following are acceptable:

    ```python
    async def method_ellipsis(self) -> None:
        ...

    async def method_pass(self) -> None:
        pass

    async def method_docstring(self) -> None:
        '''This is a docstring'''
    ```

    Args:
        func: The function or method being checked

    Returns:
        bool: Whether the function is empty and should be auto-executed
    """
    src = dedent(getsource(func))
    tree = parse(src)

    assert isinstance(tree.body[0], (FunctionDef, AsyncFunctionDef))
    body = tree.body[0].body

    return len(body) == 1 and (
        (
            isinstance(body[0], Expr)
            and isinstance(body[0].value, Constant)
            and (
                body[0].value.value is Ellipsis
                or (
                    isinstance(body[0].value.value, str)
                    and cleandoc(body[0].value.value)
                    == cleandoc(getdoc(func) or "")
                )
            )
        )
        or isinstance(body[0], Pass)
    )



import re

from mayim.exception import MayimError

DOLLAR_KEYWORD = re.compile(r"(\$([a-z][a-z0-9_]*))")
DOLLAR_POSITIONAL = re.compile(r"(\$(\d+))")


def convert_sql_params(
    query: str, positional_sub: str = r"%s", keyword_sub: str = r"%(\2)s"
) -> str:
    matches = 0
    if DOLLAR_KEYWORD.search(query):
        matches += 1
        query = DOLLAR_KEYWORD.sub(keyword_sub, query, 0)
    if DOLLAR_POSITIONAL.search(query):
        matches += 1
        query = DOLLAR_POSITIONAL.sub(positional_sub, query, 0)
    if matches > 1:
        raise MayimError(f"Could not properly convert SQL params {matches}")
    return query



from collections import defaultdict
from inspect import isclass
from typing import DefaultDict

from mayim.registry import Registry
from mayim.sql.executor import SQLExecutor


class SQLCounterMixin(SQLExecutor):
    def __init__(self, *args, **kwargs) -> None:
        self.reset()
        super().__init__(*args, **kwargs)

    def reset(self):
        self._counter = defaultdict(int)

    def _run_sql(self, *args, name: str = "", **kwargs):
        if name and self.is_query_name(name):
            query_type, _ = name.split("_", 1)
        else:
            query_type = "unknown"
        self._counter[query_type] += 1
        kwargs["name"] = name
        return super()._run_sql(*args, **kwargs)


class SQLStatisticsMiddleware:
    def __init__(self, app, logger):
        self.app = app
        self.logger = logger

    async def __call__(self, scope, receive, send):
        if scope["type"] == "http":
            setup_query_counter()
        response = await self.app(scope, receive, send)
        if scope["type"] == "http":
            log_statistics_report(self.logger)

        return response


def display_statistics(counters, executors) -> bool:
    if isinstance(counters, bool):
        return counters

    def _is_sql_counter(executor):
        executor_class = executor if isclass(executor) else executor.__class__
        return SQLCounterMixin in executor_class.mro()

    return any(_is_sql_counter(executor) for executor in executors)


def setup_query_counter(*_, **__):
    registry = Registry()
    for executor in registry.values():
        if hasattr(executor, "reset"):
            executor.reset()


def log_statistics_report(logger, *_):
    COLUMN_SIZE = 6
    registry = Registry()
    statistics = [
        dict(executor._counter)
        for executor in registry.values()
        if hasattr(executor, "_counter")
    ]
    keys = list(
        sorted(
            {
                key.rjust(COLUMN_SIZE)
                for executor in registry.values()
                if hasattr(executor, "_counter")
                for key in executor._counter.keys()
            }
        )
    )
    max_executor_name = max(map(len, registry.keys()))
    headers = " | ".join([" " * max_executor_name, *keys])
    row_data = [
        " | ".join(
            [
                name.rjust(max_executor_name),
                *[
                    str(executor._counter.get(key, "-")).rjust(COLUMN_SIZE)
                    for key in keys
                ],
            ]
        )
        for name, executor in sorted(registry.items(), key=lambda x: x[0])
        if hasattr(executor, "_counter")
    ]
    if not row_data:
        logger.warning("No executor counters found")
        return
    rows = "\n".join(row_data)
    total_values: DefaultDict[str, int] = defaultdict(int)
    for stats in statistics:
        for key, value in stats.items():
            total_values[key] += value
    divider = "=" * len(row_data[0])
    totals = " | ".join(
        [
            "TOTALS".rjust(max_executor_name),
            *[
                str(total_values.get(key, "-")).rjust(COLUMN_SIZE)
                for key in keys
            ],
        ]
    )
    title = "QUERY COUNTERS".center(len(divider))

    logger.info(
        f"SQL Statistics Report\n\n{title}\n\n{headers}\n"
        f"{rows}\n{divider}\n{totals}\n\n"
    )



from inspect import cleandoc

from mayim.base.hydrator import Hydrator
from mayim.registry import LazyHydratorRegistry, LazyQueryRegistry, Registry


def query(query: str):
    """Convenience decorator to supply a query to an executor method without
    loading if from a source file.

    Example:

    ```python
    from mayim import PostgresExecutor, query

    class ItemExecutor(PostgresExecutor):
        @query(
            '''
            SELECT *
            FROM items
            WHERE item_id = $item_id;
            '''
        )
        async def select_item(self, item_id: int) -> Item:
            ...
    ```

    Args:
        query (str): The query
    """

    def decorator(f):
        *_, class_name, method_name = f.__qualname__.rsplit(".", 2)
        LazyQueryRegistry.add(class_name, method_name, cleandoc(query))
        return f

    return decorator


def hydrator(hydrator: Hydrator):
    """Convenience decorator to supply a specific hydrator to an
    executor method.

    Example:

    ```python
    from mayim import Mayim, Executor, Hydrator, hydrator

    class HydratorA(Hydrator):
        ...

    class HydratorB(Hydrator):
        ...

    class SomeExecuto(Executor):
        async def select_a(...) -> Something:
            ...

        @hydrator(HydratorB())
        async def select_b(...) -> Something:
            ...

    Mayim(executors=[SomeExecutor(hydrator=HydratorA())])
    ```

    Args:
        hydrator (Hydrator): The hydrator
    """

    def decorator(f):
        *_, class_name, method_name = f.__qualname__.rsplit(".", 2)
        LazyHydratorRegistry.add(class_name, method_name, hydrator)
        return f

    return decorator


def register(cls):
    """Convenience decorator to preregister an executor

    Example:

    ```python
    from mayim import PostgresExecutor, register

    @register
    class MyExecutor(PostgresExecutor):
        async def select_something(self) -> Something:
            ...
    ```
    """
    Registry().register(cls)
    return cls



from __future__ import annotations

from typing import Any, Dict, Optional, Sequence

from mayim.sql.postgres.query import PostgresQuery

from ..executor import SQLExecutor

try:
    from psycopg.rows import dict_row

    POSTGRES_ENABLED = True
except ModuleNotFoundError:
    POSTGRES_ENABLED = False


class PostgresExecutor(SQLExecutor):
    """Executor for interfacing with a Postgres database"""

    ENABLED = POSTGRES_ENABLED
    QUERY_CLASS = PostgresQuery

    async def _run_sql(
        self,
        query: str,
        name: str = "",
        as_list: bool = False,
        no_result: bool = False,
        posargs: Optional[Sequence[Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        method_name = self._get_method(as_list=as_list)
        async with self.pool.connection() as conn:
            exec_values = list(posargs) if posargs else params
            cursor = await conn.execute(query, exec_values)
            if no_result:
                return None
            cursor.row_factory = dict_row
            raw = await getattr(cursor, method_name)()
            return raw



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import namedtuple
from contextvars import ContextVar
from typing import Any, Optional, Set, Type
from urllib.parse import urlparse

from mayim.exception import MayimError
from mayim.registry import InterfaceRegistry

UrlMapping = namedtuple("UrlMapping", ("key", "cast"))


URLPARSE_MAPPING = {
    "hostname": UrlMapping("_host", str),
    "username": UrlMapping("_user", str),
    "password": UrlMapping("_password", str),
    "port": UrlMapping("_port", int),
    "path": UrlMapping("_db", lambda value: value.replace("/", "")),
}


class BaseInterface(ABC):
    scheme = "dummy"
    registered_interfaces: Set[Type[BaseInterface]] = set()

    def __init_subclass__(cls) -> None:
        BaseInterface.registered_interfaces.add(cls)

    @abstractmethod
    def _setup_pool(self):
        ...

    @abstractmethod
    async def open(self):
        ...

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    def connection(self, timeout: Optional[float] = None):
        ...

    def __init__(
        self,
        dsn: Optional[str] = None,
        host: Optional[int] = None,
        port: Optional[int] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[int] = None,
    ) -> None:
        """DB class initialization.

        Args:
            dsn (str, optional): DB data source name
            host (str, optional): DB address URL or IP
            port (int, optional): DB port. Defaults to 6379
            password (str, optional): DB password
            db (int, optional): DB db. Defaults to 1
        """

        if dsn and host:
            raise MayimError("Cannot connect to DB using host and dsn")

        if not dsn:
            if port and (
                not isinstance(port, int) or port not in range(0, 65536)
            ):
                raise MayimError(
                    "port: must be an integer between 0 and 65535"
                )

            if host and (not isinstance(host, str) or not len(host) > 0):
                raise MayimError(
                    "host: must be a string at least 1 character long"
                )

        if password is not None and (
            not len(password) > 0 or not isinstance(password, str)
        ):
            raise MayimError(
                "password: must be a string at least 1 character long"
            )

        self._dsn = dsn
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._db = db
        self._full_dsn: Optional[str] = None
        self._connection: ContextVar[Any] = ContextVar(
            "connection", default=None
        )
        self._transaction: ContextVar[bool] = ContextVar(
            "transaction", default=False
        )
        self._commit: ContextVar[bool] = ContextVar("commit", default=True)

        self._populate_connection_args()
        self._populate_dsn()
        self._setup_pool()
        InterfaceRegistry.add(self)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.dsn}>"

    def _populate_connection_args(self):
        dsn = self.dsn or ""
        if dsn:
            parts = urlparse(dsn)
            for key, mapping in URLPARSE_MAPPING.items():
                if not getattr(self, mapping.key):
                    value = getattr(parts, key, None)  # or TODO: make defaults
                    setattr(self, mapping.key, mapping.cast(value))

    def _populate_dsn(self):
        self._dsn = (
            (
                f"{self.scheme}://{self.user}:...@"
                f"{self.host}:{self.port}/{self.db}"
            )
            if self.password
            else (
                f"{self.scheme}://{self.user}@"
                f"{self.host}:{self.port}/{self.db}"
            )
        )
        self._full_dsn = (
            (
                f"{self.scheme}://{self.user}:{self.password}@"
                f"{self.host}:{self.port}/{self.db}"
            )
            if self.password
            else self.dsn
        )

    @property
    def dsn(self):
        return self._dsn

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def user(self):
        return self._user

    @property
    def password(self):
        return self._password

    @property
    def db(self):
        return self._db

    @property
    def full_dsn(self):
        return self._full_dsn

    def existing_connection(self):
        return self._connection.get()

    def in_transaction(self) -> bool:
        return self._transaction.get()

    def do_commit(self) -> bool:
        return self._commit.get()



from __future__ import annotations

from typing import Any, Dict, Optional, Sequence

from mayim.sql.mysql.query import MysqlQuery

from ..executor import SQLExecutor

try:
    from asyncmy.cursors import DictCursor

    MYSQL_ENABLED = True
except ModuleNotFoundError:
    MYSQL_ENABLED = False


class MysqlExecutor(SQLExecutor):
    """Executor for interfacing with a MySQL database"""

    ENABLED = MYSQL_ENABLED
    QUERY_CLASS = MysqlQuery

    async def _run_sql(
        self,
        query: str,
        name: str = "",
        as_list: bool = False,
        no_result: bool = False,
        posargs: Optional[Sequence[Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        method_name = self._get_method(as_list=as_list)
        async with self.pool.connection() as conn:
            async with conn.cursor(cursor=DictCursor) as cursor:
                exec_values = list(posargs) if posargs else params
                await cursor.execute(query, exec_values)
                if no_result:
                    return None
                raw = await getattr(cursor, method_name)()
                return raw



from __future__ import annotations

from logging import INFO, Logger, basicConfig, getLogger
from typing import Optional, Sequence, Type, Union

from mayim import Executor, Hydrator, Mayim
from mayim.base.interface import BaseInterface
from mayim.exception import MayimError
from mayim.extension.statistics import (
    SQLStatisticsMiddleware,
    display_statistics,
)
from mayim.registry import InterfaceRegistry, Registry

try:
    from starlette.applications import Starlette

    STARLETTE_INSTALLED = True
except ModuleNotFoundError:
    STARLETTE_INSTALLED = False
    Starlette = type("Starlette", (), {})  # type: ignore


class Default:
    ...


_default = Default()


class StarletteMayimExtension:
    def __init__(
        self,
        *,
        executors: Optional[Sequence[Union[Type[Executor], Executor]]] = None,
        dsn: str = "",
        hydrator: Optional[Hydrator] = None,
        pool: Optional[BaseInterface] = None,
        app: Optional[Starlette] = None,
        counters: Union[Default, bool] = _default,
    ):
        if not STARLETTE_INSTALLED:
            raise MayimError(
                "Could not locate Starlette. It must be installed to use "
                "StarletteMayimExtension. Try: pip install starlette"
            )
        self.executors = executors or []
        for executor in self.executors:
            Registry().register(executor)
        self.mayim_kwargs = {
            "dsn": dsn,
            "hydrator": hydrator,
            "pool": pool,
        }
        self.counters = counters
        if app is not None:
            self.init_app(app)

    def init_app(
        self, app: Starlette, logger: Optional[Logger] = None
    ) -> None:
        async def startup():
            Mayim(executors=self.executors, **self.mayim_kwargs)
            for interface in InterfaceRegistry():
                await interface.open()

        async def shutdown():
            for interface in InterfaceRegistry():
                await interface.close()

        app.add_event_handler("startup", startup)
        app.add_event_handler("shutdown", shutdown)

        if display_statistics(self.counters, self.executors):
            if logger is None:
                basicConfig(level=INFO)
                logger = getLogger("mayim")
            app.add_middleware(SQLStatisticsMiddleware, logger=logger)
